import os

from pythonjsonlogger.jsonlogger import JsonFormatter, merge_record_extra

bind = ["0.0.0.0:8000"]
forwarded_allow_ips = "*"
workers = 5
reload = os.getenv("PYQUOCCA_RELOAD", "").lower() not in ["", "false", "0", "f", "no"]


# Based on https://stackoverflow.com/a/70511781
class GunicornJsonFormatter(JsonFormatter):
    def add_fields(self, log_record, record, message_dict):
        """
        This method allows us to inject gunicorn's args as fields for the formatter
        """
        super(GunicornJsonFormatter, self).add_fields(log_record, record, message_dict)
        for field in self._required_fields:
            if record.args and field in record.args:
                if field in self.rename_fields:
                    log_record[self.rename_fields[field]] = record.args.get(field)  # type: ignore
                else:
                    log_record[field] = record.args.get(field)  # type: ignore


logconfig_dict = {
    "version": 1,
    "formatters": {
        "json": {
            "()": "pyquocca.gunicorn.GunicornJsonFormatter",
            "format": "%(M)s %(t)s %(r)s %(s)s %(p)s %(a)s %({host}i)s %({x-mtls-full-name}i)s %({x-forwarded-for}i)s %({x-mtls-username}i)s %({x-mtls-staff}i)s %({x-mtls-impersonated-by}i)s",
            "datefmt": "%Y-%m-%dT%H:%M:%S%z",
            "rename_fields": {
                "M": "duration",
                "a": "user_agent",
                "t": "time",
                "{host}i": "host",
                "{x-forwarded-for}i": "ip",
                "s": "status",
                "r": "request",
                "p": "process_id",
                "{x-mtls-full-name}i": "full_name",
                "{x-mtls-username}i": "user",
                "{x-mtls-staff}i": "is_staff",
                "{x-mtls-impersonated-by}i": "impersonated_by",
            },
        }
    },
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "level": "INFO",
            "formatter": "json",
            "stream": "ext://sys.stdout",
        }
    },
    "loggers": {"gunicorn.access": {"level": "INFO", "handlers": ["console"]}},
}



from typing import Optional

from flask import Request
from flask import request as flask_request

USERNAME_HEADER = "X-mTLS-Username"
STAFF_HEADER = "X-mTLS-Staff"
FIRST_NAME_HEADER = "X-mTLS-First-Name"
LAST_NAME_HEADER = "X-mTLS-Last-Name"
FULL_NAME_HEADER = "X-mTLS-Full-Name"
IMPERSONATED_BY_HEADER = "X-mTLS-Impersonated-By"

# Would be nice to have a system to differentiate bots from regular staff.
# Maybe a third CA?
XSSBOT_USERNAME = "XSSBot"


def get_username(request: Optional[Request] = None) -> Optional[str]:
    """Returns the username from the current request."""
    r = request if request is not None else flask_request
    return r.headers.get(USERNAME_HEADER)


def get_impersonator(request: Optional[Request] = None) -> Optional[str]:
    """A user may be being impersonated by a staff user. This functions return the impersonator's username."""
    r = request if request is not None else flask_request
    return r.headers.get(IMPERSONATED_BY_HEADER)


def is_staff(request: Optional[Request] = None) -> bool:
    """Returns whether or not the request being made is made by a staff user."""
    r = request if request is not None else flask_request
    return r.headers.get(STAFF_HEADER, "false") == "true"


def is_xssbot(request: Optional[Request] = None) -> bool:
    """Determines if the current request is being made by XSSBot, either directly or via impersonation."""
    r = request if request is not None else flask_request
    return (is_staff(r) and get_username(r) == XSSBOT_USERNAME) or (
        get_impersonator() == XSSBOT_USERNAME
    )


def get_first_name(request: Optional[Request] = None) -> Optional[str]:
    """Returns the first name of the current user if available, then falls back to username."""
    r = request if request is not None else flask_request
    return r.headers.get(FIRST_NAME_HEADER, get_username(request))


def get_last_name(request: Optional[Request] = None) -> Optional[str]:
    """Returns the last name of the current user if available, then falls back to username."""
    r = request if request is not None else flask_request
    return r.headers.get(LAST_NAME_HEADER, get_username(request))


def get_full_name(request: Optional[Request] = None) -> Optional[str]:
    """Returns the full name of the current user if available, then falls back to username."""
    r = request if request is not None else flask_request
    return r.headers.get(FULL_NAME_HEADER, get_username(request))



import os
import re
from typing import Any, Callable, Optional, TypeVar, Union, cast

from flask import Flask, g
from pymysql import Connection
from pymysql.cursors import DictCursor

_NAME_TO_ENV = re.compile(r"[^A-Z0-9_]")


def connect(name: str):
    """Connects to a MySQL database resource using provided environment variables (e.g. `MYSQL_{name}_HOST`).

    :param name: name of MySQL resource as defined in chal.yaml
    :type name: str
    :return: pymysql connection object
    :rtype: pymysql.Connection
    """

    env_name = _NAME_TO_ENV.sub("", name.upper().replace("-", "_"))

    host = os.getenv(f"MYSQL_{env_name}_HOST")
    database = os.getenv(f"MYSQL_{env_name}_DB")
    user = os.getenv(f"MYSQL_{env_name}_USER")
    password = os.getenv(f"MYSQL_{env_name}_PASS")

    assert (
        host is not None
        and database is not None
        and user is not None
        and password is not None
    ), f"Environment variables for MySQL resource `{name}` not found (e.g. `MYSQL_{env_name}_HOST`)."

    return Connection(
        host=host,
        database=database,
        user=user,
        password=password,
        cursorclass=DictCursor,
    )


T = TypeVar("T")


def execute(
    connection: Connection,
    sql: str,
    values: Optional[Union[tuple, list]] = None,
    fetcher: Callable[[DictCursor], T] = lambda cursor: None,
):
    """Executes an SQL query against a database connection.

    :param connection: pymysql connection
    :type connection: Connection
    :param sql: SQL query to execute
    :type sql: str
    :param values: parametrised values, defaults to None
    :type values: tuple or list of values, optional
    :param fetcher: callback to execute with cursor to fetch results if any, defaults to `lambda cursor: None`
    :type fetcher: function taking a cursor and returning some result type, optional
    :return: return value of fetcher
    :rtype: return type of fetcher
    """
    with connection.cursor() as cursor:
        cursor.execute(sql, values)
        return fetcher(cursor)


def fetch_one(connection: Connection, sql: str, values=None):
    """Executes an SQL query and calls `cursor.fetchone()` automatically."""
    return execute(connection, sql, values, fetcher=lambda cursor: cursor.fetchone())


def fetch_all(connection: Connection, sql: str, values=None):
    """Executes an SQL query and calls `cursor.fetchall()` automatically."""
    return execute(
        connection, sql, values, fetcher=lambda cursor: list(cursor.fetchall())
    )


class FlaskMySQL:
    """Flask extension to add basic transaction usage to requests. Each request gets a new connection but
    automatically commits or rolls back (if there is an unhandled exception) the entire transaction at
    the end of the request.
    """

    def __init__(self, name: str, app: Optional[Flask] = None):
        self.name = name

        if app is not None:
            self.init_app(app)

    def init_app(self, app: Flask):
        app.teardown_request(self._teardown_request)

    def _get_connections_dict(self):
        # g._connections is a dictionary of database names to connections.
        try:
            assert type(g._connections) is dict
            return g._connections
        except (AttributeError, AssertionError):
            g._connections = {}
            return g._connections

    def _get_connection(self):
        connections = self._get_connections_dict()
        if self.name not in connections:
            return None
        return cast(
            Connection,
            connections[self.name],
        )

    def _get_or_create_connection(self):
        db = self._get_connection()
        if db is None:
            db = connect(self.name)
            self._get_connections_dict()[self.name] = db
        return db

    def _teardown_request(self, exception: Optional[BaseException]):
        db = self._get_connection()

        if db is not None:
            if exception is None:
                db.commit()
                db.close()
            else:
                db.close()

            del self._get_connections_dict()[self.name]

    def execute(
        self,
        sql: str,
        values=None,
        fetcher: Callable[[DictCursor], T] = lambda cursor: None,
    ):
        """Executes an SQL query.

        :param sql: SQL query to execute
        :type sql: str
        :param values: parametrised values, defaults to None
        :type values: tuple or list of values, optional
        :param fetcher: callback to execute with cursor to fetch results if any, defaults to `lambda cursor: None`
        :type fetcher: function taking a cursor and returning some result type, optional
        :return: return value of fetcher
        :rtype: return type of fetcher
        """
        return execute(self._get_or_create_connection(), sql, values, fetcher)

    def fetch_one(self, sql: str, values=None):
        """Executes an SQL query and calls `cursor.fetchone()` automatically."""
        return fetch_one(self._get_or_create_connection(), sql, values)

    def fetch_all(self, sql: str, values=None) -> list[dict[str, Any]]:
        """Executes an SQL query and calls `cursor.fetchall()` automatically."""
        return fetch_all(self._get_or_create_connection(), sql, values)



import logging
import os
from typing import Optional

import requests

from .header_utils import get_username

logger = logging.getLogger(__name__)

FLAGANIZER_URL = os.getenv("FLAGANIZER_URL")
dummy = FLAGANIZER_URL is None
if dummy:
    logger.warn(
        "FLAGANIZER_URL environment variable not found, generating dummy flags."
    )
else:
    FLAGANIZER_URL = FLAGANIZER_URL.removesuffix("/")


class FlaganizerException(Exception):
    pass


class MissingTokenError(FlaganizerException):
    def __init__(self, key: str):
        super().__init__(f"This service does not have access to the token for {key}.")


def _get_token(key: str) -> str:
    try:
        with open(f"/run/secrets/flaganizer/{key}/token") as f:
            return f.read()
    except FileNotFoundError as e:
        raise MissingTokenError(key) from e


def get_flag(key: str, username: Optional[str] = None) -> str:
    if dummy:
        return f"FLAG{{{key}}}"

    if username is None:
        username = get_username()
        if username is None:
            raise ValueError("Must be running with a user or request.")

    assert FLAGANIZER_URL is not None, "flaganizer url missing"
    r = requests.get(
        f"{FLAGANIZER_URL}/generate",
        params={"username": username},
        headers={
            "Authorization": f"Bearer {_get_token(key)}",
            "User-Agent": "pyquocca",
        },
    )
    if r.status_code == 200:
        return r.json().get("flag")
    else:
        raise FlaganizerException(r.json().get("msg"))



import os
from typing import Any, Literal, Optional

import requests
from typing_extensions import NotRequired, TypedDict

from .header_utils import get_username, is_staff

XSSBOT_URL = os.getenv("XSSBOT_URL")


class XSSBotException(Exception):
    pass


class MissingURLError(XSSBotException):
    def __init__(self):
        super().__init__(
            f"Missing XSSBOT_URL environment variable for accessing XSSBot."
        )


class BadRequestError(XSSBotException):
    def __init__(self, response: Any):
        self.response = response
        super().__init__(f"XSSBot returned a 400 Bad Request response: {response!r}")


class Cookie(TypedDict):
    name: str
    value: str
    domain: str
    path: NotRequired[str]
    sameSite: NotRequired[Literal["Strict", "Lax", "None"]]
    httpOnly: NotRequired[bool]
    secure: NotRequired[bool]


class Options(TypedDict):
    userAgent: NotRequired[str]
    mtlsDomains: NotRequired[list[str]]


def visit(
    url: str,
    /,
    cookies: list[Cookie],
    *,
    spoof_mtls=True,
    timeout: Optional[int] = None,
    network_timeout: Optional[int] = None,
    options: Options = {},
):
    """Tells XSSBot to visit a page using a headless browser to check if a user's XSS payload has worked.

    :param url: URL of site/page to visit
    :type url: str
    :param cookies: list of cookies to set for the visit
    :type cookies: list[Cookie]
    :param spoof_mtls: sets whether or not the current user's mTLS identity will be spoofed if available, defaults to True
    :type spoof_mtls: bool, optional
    :param timeout: sets the timeout for the entire visit, default is decided by XSSBot
    :type timeout: Optional[int], optional
    :param network_timeout: sets the timeout for network calls, default is decided by XSSBot
    :type network_timeout: Optional[int], optional
    :raises MissingURLError: if the XSSBOT_URL environment variable isn't set
    :raises BadRequestError: if XSSBot responds with a 400 Bad Request error
    :raises XSSBotException: if XSSBot responds with any other non-200 response code
    """
    if XSSBOT_URL is None:
        raise MissingURLError()

    username = get_username()
    if spoof_mtls and username is not None:
        mtls_user = {"username": username, "isStaff": is_staff()}
    else:
        mtls_user = None

    data = {"url": url, "cookies": cookies, **options}

    if timeout:
        data["timeout"] = timeout
    if network_timeout:
        data["networkTimeout"] = network_timeout
    if mtls_user:
        data["mtlsUser"] = mtls_user

    r = requests.post(
        f"{XSSBOT_URL}/visit",
        headers={
            "User-Agent": "pyquocca",
        },
        json=data,
    )

    if r.status_code == 204:
        return
    elif r.status_code == 400:
        raise BadRequestError(r.json())
    else:
        raise XSSBotException(r.content)
